@@ -1,4 +1,5 @@
 # -*- coding: utf-8 -*-
+from six import raise_from
 import errno
 import io
 import logging
@@ -8,7 +9,6 @@ import socket
 from hashlib import md5
 
 import boto3
-import boto3.compat
 from botocore.client import Config
 from botocore.exceptions import ClientError, ParamValidationError
 
@@ -28,15 +28,11 @@ except ImportError:
     )
 
 try:
-    FileNotFoundError
+    FileNotFoundError = FileNotFoundError
 except NameError:
     class FileNotFoundError(IOError):
         pass
 
-
-from six import raise_from
-
-
 # py2 has no ConnectionError only OSError with different error number for each
 # error so we need to catch OSError and compare errno to the following
 # error numbers - same as catching ConnectionError on py3
@@ -156,7 +152,7 @@ class S3FileSystem(object):
     _singleton = [None]
     connect_timeout = 5
     read_timeout = 15
-    default_block_size = 5 * 2**20
+    default_block_size = 5 * 2 ** 20
 
     def __init__(self, anon=False, key=None, secret=None, token=None,
                  use_ssl=True, client_kwargs=None, requester_pays=False,
@@ -228,8 +224,8 @@ class S3FileSystem(object):
             Whether to use cached filelists, if already read
         """
         anon, key, secret, kwargs, ckwargs, token, ssl = (
-              self.anon, self.key, self.secret, self.kwargs,
-              self.client_kwargs, self.token, self.use_ssl)
+            self.anon, self.key, self.secret, self.kwargs,
+            self.client_kwargs, self.token, self.use_ssl)
 
         # Include the current PID in the connection key so that different
         # SSL connections are made for each process.
@@ -356,7 +352,7 @@ class S3FileSystem(object):
             return fdesc
         return io.TextIOWrapper(fdesc, encoding=encoding)
 
-    def _lsdir(self, path, refresh=False):
+    def _lsdir(self, path, refresh=False, max_items=None):
         if path.startswith('s3://'):
             path = path[len('s3://'):]
         path = path.rstrip('/')
@@ -365,8 +361,11 @@ class S3FileSystem(object):
         if path not in self.dirs or refresh:
             try:
                 pag = self.s3.get_paginator('list_objects_v2')
+                config = {}
+                if max_items is not None:
+                    config.update(MaxItems=max_items, PageSize=2 * max_items)
                 it = pag.paginate(Bucket=bucket, Prefix=prefix, Delimiter='/',
-                                  **self.req_kw)
+                                  PaginationConfig=config, **self.req_kw)
                 files = []
                 dirs = []
                 for i in it:
@@ -374,8 +373,7 @@ class S3FileSystem(object):
                     files.extend(i.get('Contents', []))
                 if dirs:
                     files.extend([{'Key': l['Prefix'][:-1], 'Size': 0,
-                                  'StorageClass': "DIRECTORY"} for l in dirs])
-                files = [f for f in files if len(f['Key']) > len(prefix)]
+                                   'StorageClass': "DIRECTORY"} for l in dirs])
                 for f in files:
                     f['Key'] = '/'.join([bucket, f['Key']])
             except ClientError as e:
@@ -383,7 +381,8 @@ class S3FileSystem(object):
                 if is_permission_error(e):
                     raise
                 files = []
-
+            if max_items is not None:
+                return files
             self.dirs[path] = files
         return self.dirs[path]
 
@@ -452,6 +451,45 @@ class S3FileSystem(object):
         else:
             return [f['Key'] for f in files]
 
+    def isdir(self, path, refresh=False):
+        """ Check if path points to a directory.
+
+        Check is cached unless `refresh=True`.
+
+        Parameters
+        ----------
+        path : string/bytes
+            location to check
+        refresh : bool
+            If true, don't look in the info cache
+        """
+        if path.startswith('s3://'):
+            path = path[len('s3://'):]
+        path = path.rstrip('/')
+        if not path:
+            return True
+        if not refresh:
+            parent = path.rsplit('/', 1)[0]
+            if parent in self.dirs:
+                return any(f['Key'] == path and f['StorageClass'] == 'DIRECTORY'
+                           for f in self.dirs[parent])
+        return bool(self._lsdir(path, refresh=refresh, max_items=1))
+
+    def isfile(self, path, refresh=False):
+        """ Check if path points to a file.
+
+        Check is cached unless `refresh=True`.
+
+        Parameters
+        ----------
+        path : string/bytes
+            location to check
+        refresh : bool
+            If true, don't look in the info cache
+        """
+        return not raises(FileNotFoundError,
+                          lambda: self.info(path.rstrip('/'), refresh=refresh))
+
     def info(self, path, version_id=None, refresh=False, **kwargs):
         """ Detail on the specific file pointed to by path.
 
@@ -465,28 +503,26 @@ class S3FileSystem(object):
         refresh : bool
             If true, don't look in the info cache
         """
-        parent = path.rsplit('/', 1)[0]
-
+        if path.startswith('s3://'):
+            path = path[len('s3://'):]
+        path = path.rstrip('/')
         if not refresh:
-            if path in self.dirs:
-                files = self.dirs[path]
-                if len(files) == 1:
-                    return files[0]
-            elif parent in self.dirs:
+            parent = path.rsplit('/', 1)[0]
+            if parent != path and parent in self.dirs:
                 for f in self.dirs[parent]:
                     if f['Key'] == path:
                         return f
 
+        if version_id is not None:
+            if not self.version_aware:
+                raise ValueError("version_id cannot be specified if the "
+                                 "filesystem is not version aware")
+            kwargs['VersionId'] = version_id
         try:
             bucket, key = split_path(path)
-            if version_id is not None:
-                if not self.version_aware:
-                    raise ValueError("version_id cannot be specified if the "
-                                     "filesystem is not version aware")
-                kwargs['VersionId'] = version_id
             out = self._call_s3(self.s3.head_object, kwargs, Bucket=bucket,
                                 Key=key, **self.req_kw)
-            out = {
+            return {
                 'ETag': out['ETag'],
                 'Key': '/'.join([bucket, key]),
                 'LastModified': out['LastModified'],
@@ -494,7 +530,6 @@ class S3FileSystem(object):
                 'StorageClass': "STANDARD",
                 'VersionId': out.get('VersionId')
             }
-            return out
         except (ClientError, ParamValidationError) as e:
             logger.debug("Failed to head path %s", path, exc_info=True)
             raise_from(FileNotFoundError(path), e)
@@ -648,7 +683,7 @@ class S3FileSystem(object):
             Key=key,
             Metadata=metadata,
             MetadataDirective='REPLACE',
-            )
+        )
 
         # refresh metadata
         self._metadata_cache[path] = metadata
@@ -693,11 +728,18 @@ class S3FileSystem(object):
             self._call_s3(self.s3.put_bucket_acl,
                           kwargs, Bucket=bucket, ACL=acl)
 
-    def glob(self, path):
+    def glob(self, path, refresh=False):
         """
         Find files by glob-matching.
 
         Note that the bucket part of the path must not contain a "*"
+
+        Parameters
+        ----------
+        path : string/bytes
+            location at which to list files with glob characters
+        refresh : bool (=False)
+            if False, look in local cache for files
         """
         path0 = path
         if path.startswith('s3://'):
@@ -712,7 +754,7 @@ class S3FileSystem(object):
         # must be a / between the bucket name and the *.
         star_pos = path.find('*')
         root = path if star_pos == -1 else path[:path[:star_pos].rindex('/') + 1]
-        allfiles = self.walk(root, directories=True)
+        allfiles = self.walk(root, refresh=refresh, directories=True)
 
         # Translate the glob pattern into regex (similar to `fnmatch`).
         regex_text = '('
@@ -729,8 +771,8 @@ class S3FileSystem(object):
 
         seen = set()
         return [p for f in allfiles
-                      for m in [regex_pattern.match(f)] if m
-                          for p in [m.group(0)] if not (p in seen or seen.add(p))]
+                for m in [regex_pattern.match(f)] if m
+                for p in [m.group(0)] if not (p in seen or seen.add(p))]
 
     def du(self, path, total=False, deep=False, **kwargs):
         """ Bytes in keys at path """
@@ -818,6 +860,7 @@ class S3FileSystem(object):
             region in which the bucket should be created
         """
         acl = acl or self.s3_additional_kwargs.get('ACL', '')
+        path = path.rstrip('/') + '/'
         self.touch(path, acl=acl, **kwargs)
 
     def rmdir(self, path, **kwargs):
@@ -856,7 +899,7 @@ class S3FileSystem(object):
             kwargs,
             Bucket=bucket,
             Key=key
-            )
+        )
         out = [self._call_s3(
             self.s3.upload_part_copy,
             kwargs,
@@ -880,7 +923,7 @@ class S3FileSystem(object):
                 self.s3.copy_object,
                 kwargs,
                 Bucket=buc2, Key=key2, CopySource='/'.join([buc1, key1])
-                )
+            )
         except (ClientError, ParamValidationError) as e:
             raise_from(IOError('Copy failed', (path1, path2)), e)
 
@@ -926,7 +969,7 @@ class S3FileSystem(object):
         bucket = buckets.pop()
         if len(pathlist) > 1000:
             for i in range((len(pathlist) // 1000) + 1):
-                self.bulk_delete(pathlist[i*1000:(i+1)*1000])
+                self.bulk_delete(pathlist[i * 1000:(i + 1) * 1000])
             return
         delete_keys = {'Objects': [{'Key': split_path(path)[1]} for path
                                    in pathlist]}
@@ -956,9 +999,7 @@ class S3FileSystem(object):
             raise FileNotFoundError(path)
         if recursive:
             self.invalidate_cache(path)
-            self.bulk_delete(self.walk(path), **kwargs)
-            if not self.exists(path):
-                return
+            self.bulk_delete(self.walk(path, directories=True), **kwargs)
         bucket, key = split_path(path)
         if key:
             try:
@@ -1139,7 +1180,7 @@ class S3File(object):
             self.forced = False
             if 'a' in mode and s3.exists(path):
                 self.size = s3.info(path)['Size']
-                if self.size < 5*2**20:
+                if self.size < 5 * 2 ** 20:
                     # existing file too small for multi-upload: download
                     self.write(s3.cat(path))
                 else:
@@ -1401,7 +1442,8 @@ class S3File(object):
                     self.s3.s3.create_multipart_upload,
                     Bucket=self.bucket, Key=self.key, ACL=self.acl)
             except (ClientError, ParamValidationError) as e:
-                raise_from(IOError('Initiating write failed: %s' % self.path), e)
+                raise_from(IOError('Initiating write failed: %s' %
+                                   self.path), e)
 
             while True:
                 try:
@@ -1495,7 +1537,7 @@ class S3File(object):
         raise io.UnsupportedOperation()
 
     def readinto(self, b):
-        data = self.read()
+        data = self.read(len(b))
         b[:len(data)] = data
         return len(data)
 
