@@ -6,7 +6,7 @@ import re
 import time
 import pytest
 from itertools import chain
-from s3fs.core import S3FileSystem
+from s3fs.core import S3FileSystem, FileNotFoundError
 from s3fs.utils import seek_delimiter, ignoring, SSEParams
 import moto
 import boto3
@@ -16,14 +16,14 @@ from botocore.exceptions import NoCredentialsError
 test_bucket_name = 'test'
 secure_bucket_name = 'test-secure'
 versioned_bucket_name = 'test-versioned'
-files = {'test/accounts.1.json':  (b'{"amount": 100, "name": "Alice"}\n'
-                                   b'{"amount": 200, "name": "Bob"}\n'
-                                   b'{"amount": 300, "name": "Charlie"}\n'
-                                   b'{"amount": 400, "name": "Dennis"}\n'),
-         'test/accounts.2.json':  (b'{"amount": 500, "name": "Alice"}\n'
-                                   b'{"amount": 600, "name": "Bob"}\n'
-                                   b'{"amount": 700, "name": "Charlie"}\n'
-                                   b'{"amount": 800, "name": "Dennis"}\n')}
+files = {'test/accounts.1.json': (b'{"amount": 100, "name": "Alice"}\n'
+                                  b'{"amount": 200, "name": "Bob"}\n'
+                                  b'{"amount": 300, "name": "Charlie"}\n'
+                                  b'{"amount": 400, "name": "Dennis"}\n'),
+         'test/accounts.2.json': (b'{"amount": 500, "name": "Alice"}\n'
+                                  b'{"amount": 600, "name": "Bob"}\n'
+                                  b'{"amount": 700, "name": "Charlie"}\n'
+                                  b'{"amount": 800, "name": "Dennis"}\n')}
 
 csv_files = {'2014-01-01.csv': (b'name,amount,id\n'
                                 b'Alice,100,1\n'
@@ -40,10 +40,10 @@ text_files = {'nested/file1': b'hello\n',
               'nested/nested2/file2': b'world'}
 glob_files = {'file.dat': b'',
               'filexdat': b''}
-a = test_bucket_name+'/tmp/test/a'
-b = test_bucket_name+'/tmp/test/b'
-c = test_bucket_name+'/tmp/test/c'
-d = test_bucket_name+'/tmp/test/d'
+a = test_bucket_name + '/tmp/test/a'
+b = test_bucket_name + '/tmp/test/b'
+c = test_bucket_name + '/tmp/test/c'
+d = test_bucket_name + '/tmp/test/d'
 
 
 @pytest.yield_fixture
@@ -56,12 +56,15 @@ def s3():
         client = boto3.client('s3')
         client.create_bucket(Bucket=test_bucket_name, ACL='public-read')
 
-        bucket = client.create_bucket(Bucket=versioned_bucket_name, ACL='public-read')
-        bucket_versioning = boto3.resource('s3').BucketVersioning(versioned_bucket_name)
+        bucket = client.create_bucket(
+            Bucket=versioned_bucket_name, ACL='public-read')
+        bucket_versioning = boto3.resource(
+            's3').BucketVersioning(versioned_bucket_name)
         bucket_versioning.enable()
 
         # initialize secure bucket
-        bucket = client.create_bucket(Bucket=secure_bucket_name, ACL='public-read')
+        bucket = client.create_bucket(
+            Bucket=secure_bucket_name, ACL='public-read')
         policy = json.dumps({
             "Version": "2012-10-17",
             "Id": "PutObjPolicy",
@@ -94,8 +97,10 @@ def s3():
         for flist in [files, csv_files, text_files, glob_files]:
             for f, data in flist.items():
                 try:
-                    client.delete_object(Bucket=test_bucket_name, Key=f, Body=data)
-                    client.delete_object(Bucket=secure_bucket_name, Key=f, Body=data)
+                    client.delete_object(
+                        Bucket=test_bucket_name, Key=f, Body=data)
+                    client.delete_object(
+                        Bucket=secure_bucket_name, Key=f, Body=data)
                 except:
                     pass
         for k in [a, b, c, d]:
@@ -109,7 +114,7 @@ def s3():
 
 
 def test_simple(s3):
-    data = b'a' * (10 * 2**20)
+    data = b'a' * (10 * 2 ** 20)
 
     with s3.open(a, 'wb') as f:
         f.write(data)
@@ -168,6 +173,14 @@ def test_info(s3):
     assert s3.info(a) == s3.dirs[parent][0]  # correct value
     assert id(s3.info(a)) == id(s3.dirs[parent][0])  # is object from cache
 
+    new_parent = test_bucket_name + '/foo'
+    s3.mkdir(new_parent)
+    with pytest.raises(FileNotFoundError):
+        s3.info(new_parent)
+    s3.ls(new_parent)
+    with pytest.raises(FileNotFoundError):
+        s3.info(new_parent)
+
 
 test_xattr_sample_metadata = {'test_xattr': '1'}
 
@@ -176,7 +189,8 @@ def test_xattr(s3):
     bucket, key = (test_bucket_name, 'tmp/test/xattr')
     filename = bucket + '/' + key
     body = b'aaaa'
-    public_read_acl = {'Permission': 'READ', 'Grantee': {'URI': 'http://acs.amazonaws.com/groups/global/AllUsers', 'Type': 'Group'}}
+    public_read_acl = {'Permission': 'READ', 'Grantee': {
+        'URI': 'http://acs.amazonaws.com/groups/global/AllUsers', 'Type': 'Group'}}
 
     s3.s3.put_object(Bucket=bucket, Key=key,
                      ACL='public-read',
@@ -185,13 +199,16 @@ def test_xattr(s3):
 
     # save etag for later
     etag = s3.info(filename)['ETag']
-    assert public_read_acl in s3.s3.get_object_acl(Bucket=bucket, Key=key)['Grants']
+    assert public_read_acl in s3.s3.get_object_acl(
+        Bucket=bucket, Key=key)['Grants']
 
-    assert s3.getxattr(filename, 'test_xattr') == test_xattr_sample_metadata['test_xattr']
+    assert s3.getxattr(
+        filename, 'test_xattr') == test_xattr_sample_metadata['test_xattr']
     assert s3.metadata(filename) == test_xattr_sample_metadata
 
     s3file = s3.open(filename)
-    assert s3file.getxattr('test_xattr') == test_xattr_sample_metadata['test_xattr']
+    assert s3file.getxattr(
+        'test_xattr') == test_xattr_sample_metadata['test_xattr']
     assert s3file.metadata() == test_xattr_sample_metadata
 
     s3file.setxattr(test_xattr='2')
@@ -201,7 +218,8 @@ def test_xattr(s3):
     assert s3.cat(filename) == body
 
     # check that ACL and ETag are preserved after updating metadata
-    assert public_read_acl in s3.s3.get_object_acl(Bucket=bucket, Key=key)['Grants']
+    assert public_read_acl in s3.s3.get_object_acl(
+        Bucket=bucket, Key=key)['Grants']
     assert s3.info(filename)['ETag'] == etag
 
 
@@ -231,11 +249,12 @@ def test_not_delegate():
 
 
 def test_ls(s3):
-    assert set(s3.ls('')) == {test_bucket_name, secure_bucket_name, versioned_bucket_name}
+    assert set(s3.ls('')) == {test_bucket_name,
+                              secure_bucket_name, versioned_bucket_name}
     with pytest.raises((OSError, IOError)):
         s3.ls('nonexistent')
-    fn = test_bucket_name+'/test/accounts.1.json'
-    assert fn in s3.ls(test_bucket_name+'/test')
+    fn = test_bucket_name + '/test/accounts.1.json'
+    assert fn in s3.ls(test_bucket_name + '/test')
 
 
 def test_pickle(s3):
@@ -247,15 +266,78 @@ def test_pickle(s3):
 
 
 def test_ls_touch(s3):
-    assert not s3.exists(test_bucket_name+'/tmp/test')
+    assert not s3.exists(test_bucket_name + '/tmp/test')
     s3.touch(a)
     s3.touch(b)
-    L = s3.ls(test_bucket_name+'/tmp/test', True)
+    L = s3.ls(test_bucket_name + '/tmp/test', True)
     assert {d['Key'] for d in L} == {a, b}
-    L = s3.ls(test_bucket_name+'/tmp/test', False)
+    L = s3.ls(test_bucket_name + '/tmp/test', False)
     assert set(L) == {a, b}
 
 
+def test_isfile(s3):
+    assert not s3.isfile('')
+    assert not s3.isfile('/')
+    assert not s3.isfile(test_bucket_name)
+    assert not s3.isfile(test_bucket_name + '/test')
+
+    assert not s3.isfile(test_bucket_name + '/test/foo')
+    assert s3.isfile(test_bucket_name + '/test/accounts.1.json')
+    assert s3.isfile(test_bucket_name + '/test/accounts.2.json')
+
+    assert not s3.isfile(a)
+    s3.touch(a)
+    assert s3.isfile(a)
+
+    assert not s3.isfile(b)
+    assert not s3.isfile(b + '/')
+    s3.mkdir(b)
+    assert not s3.isfile(b)
+    assert not s3.isfile(b + '/')
+
+    assert not s3.isfile(c)
+    assert not s3.isfile(c + '/')
+    s3.mkdir(c + '/')
+    assert not s3.isfile(c)
+    assert not s3.isfile(c + '/')
+
+
+def test_isdir(s3):
+    assert s3.isdir('')
+    assert s3.isdir('/')
+    assert s3.isdir(test_bucket_name)
+    assert s3.isdir(test_bucket_name + '/test')
+
+    assert not s3.isdir(test_bucket_name + '/test/foo')
+    assert not s3.isdir(test_bucket_name + '/test/accounts.1.json')
+    assert not s3.isdir(test_bucket_name + '/test/accounts.2.json')
+
+    assert not s3.isdir(a)
+    s3.touch(a)
+    assert not s3.isdir(a)
+
+    assert not s3.isdir(b)
+    assert not s3.isdir(b + '/')
+    s3.mkdir(b)
+    assert s3.isdir(b)
+    assert s3.isdir(b + '/')
+
+    assert not s3.isdir(c)
+    assert not s3.isdir(c + '/')
+    s3.mkdir(c + '/')
+    assert s3.isdir(c)
+    assert s3.isdir(c + '/')
+
+    # test cache
+    assert not s3.dirs
+    s3.ls(test_bucket_name + '/nested')
+    assert test_bucket_name + '/nested' in s3.dirs
+    assert not s3.isdir(test_bucket_name + '/nested/file1')
+    assert not s3.isdir(test_bucket_name + '/nested/file2')
+    assert s3.isdir(test_bucket_name + '/nested/nested2')
+    assert s3.isdir(test_bucket_name + '/nested/nested2/')
+
+
 def test_rm(s3):
     assert not s3.exists(a)
     s3.touch(a)
@@ -263,15 +345,15 @@ def test_rm(s3):
     s3.rm(a)
     assert not s3.exists(a)
     with pytest.raises((OSError, IOError)):
-        s3.rm(test_bucket_name+'/nonexistent')
+        s3.rm(test_bucket_name + '/nonexistent')
     with pytest.raises((OSError, IOError)):
         s3.rm('nonexistent')
-    s3.rm(test_bucket_name+'/nested', recursive=True)
-    assert not s3.exists(test_bucket_name+'/nested/nested2/file1')
+    s3.rm(test_bucket_name + '/nested', recursive=True)
+    assert not s3.exists(test_bucket_name + '/nested/nested2/file1')
 
-    #whole bucket
+    # whole bucket
     s3.rm(test_bucket_name, recursive=True)
-    assert not s3.exists(test_bucket_name+'/2014-01-01.csv')
+    assert not s3.exists(test_bucket_name + '/2014-01-01.csv')
     assert not s3.exists(test_bucket_name)
 
 
@@ -313,7 +395,7 @@ def test_mkdir_client_region_name():
         m = moto.mock_s3()
         m.start()
         s3 = S3FileSystem(anon=False, client_kwargs={"region_name":
-                                                     "eu-central-1"})
+                                                         "eu-central-1"})
         s3.mkdir(bucket)
         assert bucket in s3.ls('/')
     finally:
@@ -325,22 +407,22 @@ def test_bulk_delete(s3):
         s3.bulk_delete(['nonexistent/file'])
     with pytest.raises(ValueError):
         s3.bulk_delete(['bucket1/file', 'bucket2/file'])
-    filelist = s3.walk(test_bucket_name+'/nested')
+    filelist = s3.walk(test_bucket_name + '/nested')
     s3.bulk_delete(filelist)
-    assert not s3.exists(test_bucket_name+'/nested/nested2/file1')
+    assert not s3.exists(test_bucket_name + '/nested/nested2/file1')
 
 
 def test_anonymous_access():
     with ignoring(NoCredentialsError):
         s3 = S3FileSystem(anon=True)
         assert s3.ls('') == []
-        ## TODO: public bucket doesn't work through moto
+        # TODO: public bucket doesn't work through moto
     with pytest.raises((OSError, IOError)):
         s3.mkdir('newbucket')
 
 
 def test_s3_file_access(s3):
-    fn = test_bucket_name+'/nested/file1'
+    fn = test_bucket_name + '/nested/file1'
     data = b'hello\n'
     assert s3.cat(fn) == data
     assert s3.head(fn, 3) == data[:3]
@@ -349,82 +431,87 @@ def test_s3_file_access(s3):
 
 
 def test_s3_file_info(s3):
-    fn = test_bucket_name+'/nested/file1'
+    fn = test_bucket_name + '/nested/file1'
     data = b'hello\n'
     assert fn in s3.walk(test_bucket_name)
     assert s3.exists(fn)
-    assert not s3.exists(fn+'another')
+    assert not s3.exists(fn + 'another')
     assert s3.info(fn)['Size'] == len(data)
     with pytest.raises((OSError, IOError)):
-        s3.info(fn+'another')
+        s3.info(fn + 'another')
 
 
 def test_bucket_exists(s3):
     assert s3.exists(test_bucket_name)
-    assert not s3.exists(test_bucket_name+'x')
+    assert not s3.exists(test_bucket_name + 'x')
     s3 = S3FileSystem(anon=True)
     assert s3.exists(test_bucket_name)
-    assert not s3.exists(test_bucket_name+'x')
+    assert not s3.exists(test_bucket_name + 'x')
 
 
 def test_du(s3):
     d = s3.du(test_bucket_name, deep=True)
     assert all(isinstance(v, int) and v >= 0 for v in d.values())
-    assert test_bucket_name+'/nested/file1' in d
+    assert test_bucket_name + '/nested/file1' in d
 
-    assert s3.du(test_bucket_name + '/test/', total=True) ==\
+    assert s3.du(test_bucket_name + '/test/', total=True) == \
            sum(map(len, files.values()))
-    assert s3.du(test_bucket_name) == s3.du('s3://'+test_bucket_name)
+    assert s3.du(test_bucket_name) == s3.du('s3://' + test_bucket_name)
 
 
 def test_s3_ls(s3):
-    fn = test_bucket_name+'/nested/file1'
-    assert fn not in s3.ls(test_bucket_name+'/')
-    assert fn in s3.ls(test_bucket_name+'/nested/')
-    assert fn in s3.ls(test_bucket_name+'/nested')
-    assert s3.ls('s3://'+test_bucket_name+'/nested/') == s3.ls(test_bucket_name+'/nested')
+    fn = test_bucket_name + '/nested/file1'
+    assert fn not in s3.ls(test_bucket_name + '/')
+    assert fn in s3.ls(test_bucket_name + '/nested/')
+    assert fn in s3.ls(test_bucket_name + '/nested')
+    assert s3.ls('s3://' + test_bucket_name +
+                 '/nested/') == s3.ls(test_bucket_name + '/nested')
 
 
 def test_s3_big_ls(s3):
     for x in range(1200):
-        s3.touch(test_bucket_name+'/thousand/%i.part'%x)
+        s3.touch(test_bucket_name + '/thousand/%i.part' % x)
     assert len(s3.walk(test_bucket_name)) > 1200
-    s3.rm(test_bucket_name+'/thousand/', recursive=True)
-    assert len(s3.walk(test_bucket_name+'/thousand/')) == 0
+    s3.rm(test_bucket_name + '/thousand/', recursive=True)
+    assert len(s3.walk(test_bucket_name + '/thousand/')) == 0
 
 
 def test_s3_ls_detail(s3):
-    L = s3.ls(test_bucket_name+'/nested', detail=True)
+    L = s3.ls(test_bucket_name + '/nested', detail=True)
     assert all(isinstance(item, dict) for item in L)
 
 
 def test_s3_glob(s3):
-    fn = test_bucket_name+'/nested/file1'
-    assert fn not in s3.glob(test_bucket_name+'/')
-    assert fn not in s3.glob(test_bucket_name+'/*')
-    assert fn not in s3.glob(test_bucket_name+'/nested')
-    assert fn in s3.glob(test_bucket_name+'/nested/*')
-    assert fn in s3.glob(test_bucket_name+'/nested/file*')
-    assert fn in s3.glob(test_bucket_name+'/*/*')
+    fn = test_bucket_name + '/nested/file1'
+    assert fn not in s3.glob(test_bucket_name + '/')
+    assert fn not in s3.glob(test_bucket_name + '/*')
+    assert fn not in s3.glob(test_bucket_name + '/nested')
+    assert fn in s3.glob(test_bucket_name + '/nested/*')
+    assert fn in s3.glob(test_bucket_name + '/nested/file*')
+    assert fn in s3.glob(test_bucket_name + '/*/*')
     assert all(any(p.startswith(f + '/') or p == f
                    for p in s3.walk(test_bucket_name, directories=True))
-                       for f in s3.glob(test_bucket_name+'/nested/*'))
-    assert [test_bucket_name + '/nested/nested2'] == s3.glob(test_bucket_name + '/nested/nested2')
-    assert [ 'test/nested/nested2/file1',
-             'test/nested/nested2/file2'] == s3.glob(test_bucket_name + '/nested/nested2/*')
+               for f in s3.glob(test_bucket_name + '/nested/*'))
+    assert [test_bucket_name +
+            '/nested/nested2'] == s3.glob(test_bucket_name + '/nested/nested2')
+    assert ['test/nested/nested2/file1',
+            'test/nested/nested2/file2'] == s3.glob(test_bucket_name + '/nested/nested2/*')
 
     with pytest.raises(ValueError):
         s3.glob('*')
 
     # Make sure glob() deals with the dot character (.) correctly.
-    assert test_bucket_name+'/file.dat' in s3.glob(test_bucket_name+'/file.*')
-    assert test_bucket_name+'/filexdat' not in s3.glob(test_bucket_name+'/file.*')
+    assert test_bucket_name + '/file.dat' in s3.glob(test_bucket_name + '/file.*')
+    assert test_bucket_name + \
+           '/filexdat' not in s3.glob(test_bucket_name + '/file.*')
+
 
 def test_get_list_of_summary_objects(s3):
     L = s3.ls(test_bucket_name + '/test')
 
     assert len(L) == 2
-    assert [l.lstrip(test_bucket_name).lstrip('/') for l in sorted(L)] == sorted(list(files))
+    assert [l.lstrip(test_bucket_name).lstrip('/')
+            for l in sorted(L)] == sorted(list(files))
 
     L2 = s3.ls('s3://' + test_bucket_name + '/test')
 
@@ -441,7 +528,7 @@ def test_read_keys_from_bucket(s3):
 
 
 def test_url(s3):
-    fn = test_bucket_name+'/nested/file1'
+    fn = test_bucket_name + '/nested/file1'
     url = s3.url(fn, expires=100)
     assert 'http' in url
     assert '/nested/file1' in url
@@ -485,43 +572,44 @@ def test_bad_open(s3):
 
 
 def test_copy(s3):
-    fn = test_bucket_name+'/test/accounts.1.json'
-    s3.copy(fn, fn+'2')
-    assert s3.cat(fn) == s3.cat(fn+'2')
+    fn = test_bucket_name + '/test/accounts.1.json'
+    s3.copy(fn, fn + '2')
+    assert s3.cat(fn) == s3.cat(fn + '2')
 
 
 def test_move(s3):
-    fn = test_bucket_name+'/test/accounts.1.json'
+    fn = test_bucket_name + '/test/accounts.1.json'
     data = s3.cat(fn)
-    s3.mv(fn, fn+'2')
-    assert s3.cat(fn+'2') == data
+    s3.mv(fn, fn + '2')
+    assert s3.cat(fn + '2') == data
     assert not s3.exists(fn)
 
 
 def test_get_put(s3, tmpdir):
     test_file = str(tmpdir.join('test.json'))
 
-    s3.get(test_bucket_name+'/test/accounts.1.json', test_file)
+    s3.get(test_bucket_name + '/test/accounts.1.json', test_file)
     data = files['test/accounts.1.json']
     assert open(test_file, 'rb').read() == data
-    s3.put(test_file, test_bucket_name+'/temp')
-    assert s3.du(test_bucket_name+'/temp')[test_bucket_name+'/temp'] == len(data)
-    assert s3.cat(test_bucket_name+'/temp') == data
+    s3.put(test_file, test_bucket_name + '/temp')
+    assert s3.du(test_bucket_name +
+                 '/temp')[test_bucket_name + '/temp'] == len(data)
+    assert s3.cat(test_bucket_name + '/temp') == data
 
 
 def test_errors(s3):
     with pytest.raises((IOError, OSError)):
-        s3.open(test_bucket_name+'/tmp/test/shfoshf', 'rb')
+        s3.open(test_bucket_name + '/tmp/test/shfoshf', 'rb')
 
-    ## This is fine, no need for interleaving directories on S3
-    #with pytest.raises((IOError, OSError)):
+    # This is fine, no need for interleaving directories on S3
+    # with pytest.raises((IOError, OSError)):
     #    s3.touch('tmp/test/shfoshf/x')
 
     with pytest.raises((IOError, OSError)):
-        s3.rm(test_bucket_name+'/tmp/test/shfoshf/x')
+        s3.rm(test_bucket_name + '/tmp/test/shfoshf/x')
 
     with pytest.raises((IOError, OSError)):
-        s3.mv(test_bucket_name+'/tmp/test/shfoshf/x', 'tmp/test/shfoshf/y')
+        s3.mv(test_bucket_name + '/tmp/test/shfoshf/x', 'tmp/test/shfoshf/y')
 
     with pytest.raises((IOError, OSError)):
         s3.open('x', 'rb')
@@ -530,11 +618,11 @@ def test_errors(s3):
         s3.rm('unknown')
 
     with pytest.raises(ValueError):
-        with s3.open(test_bucket_name+'/temp', 'wb') as f:
+        with s3.open(test_bucket_name + '/temp', 'wb') as f:
             f.read()
 
     with pytest.raises(ValueError):
-        f = s3.open(test_bucket_name+'/temp', 'rb')
+        f = s3.open(test_bucket_name + '/temp', 'rb')
         f.close()
         f.read()
 
@@ -549,7 +637,7 @@ def test_errors(s3):
 
 
 def test_read_small(s3):
-    fn = test_bucket_name+'/2014-01-01.csv'
+    fn = test_bucket_name + '/2014-01-01.csv'
     with s3.open(fn, 'rb', block_size=10) as f:
         out = []
         while True:
@@ -574,7 +662,7 @@ def test_seek_delimiter(s3):
         seek_delimiter(f, b'\n', 5)
         assert f.tell() == data.index(b'\n') + 1
         f.seek(1, 1)
-        ind = data.index(b'\n') + data[data.index(b'\n')+1:].index(b'\n') + 1
+        ind = data.index(b'\n') + data[data.index(b'\n') + 1:].index(b'\n') + 1
         seek_delimiter(f, b'\n', 5)
         assert f.tell() == ind + 1
 
@@ -582,7 +670,7 @@ def test_seek_delimiter(s3):
 def test_read_s3_block(s3):
     data = files['test/accounts.1.json']
     lines = io.BytesIO(data).readlines()
-    path = test_bucket_name+'/test/accounts.1.json'
+    path = test_bucket_name + '/test/accounts.1.json'
     assert s3.read_block(path, 1, 35, b'\n') == lines[1]
     assert s3.read_block(path, 0, 30, b'\n') == lines[0]
     assert s3.read_block(path, 0, 35, b'\n') == lines[0] + lines[1]
@@ -624,11 +712,11 @@ def test_dynamic_add_rm(s3):
 
 
 def test_write_small(s3):
-    with s3.open(test_bucket_name+'/test', 'wb') as f:
+    with s3.open(test_bucket_name + '/test', 'wb') as f:
         f.write(b'hello')
-    assert s3.cat(test_bucket_name+'/test') == b'hello'
-    s3.open(test_bucket_name+'/test', 'wb').close()
-    assert s3.info(test_bucket_name+'/test')['Size'] == 0
+    assert s3.cat(test_bucket_name + '/test') == b'hello'
+    s3.open(test_bucket_name + '/test', 'wb').close()
+    assert s3.info(test_bucket_name + '/test')['Size'] == 0
 
 
 def test_write_small_secure(s3):
@@ -637,9 +725,9 @@ def test_write_small_secure(s3):
     # effectively.
     # This test is left as a placeholder in case moto eventually supports this.
     sse_params = SSEParams(server_side_encryption='aws:kms')
-    with s3.open(secure_bucket_name+'/test', 'wb', writer_kwargs=sse_params) as f:
+    with s3.open(secure_bucket_name + '/test', 'wb', writer_kwargs=sse_params) as f:
         f.write(b'hello')
-    assert s3.cat(secure_bucket_name+'/test') == b'hello'
+    assert s3.cat(secure_bucket_name + '/test') == b'hello'
     head = s3.s3.head_object(Bucket=secure_bucket_name, Key='test')
 
 
@@ -648,22 +736,22 @@ def test_write_large_secure(s3):
     mock.start()
 
     # build our own s3fs with the relevant additional kwarg
-    s3 = S3FileSystem(s3_additional_kwargs = {'ServerSideEncryption': 'AES256'})
+    s3 = S3FileSystem(s3_additional_kwargs={'ServerSideEncryption': 'AES256'})
     s3.mkdir('mybucket')
 
     with s3.open('mybucket/myfile', 'wb') as f:
-        f.write(b'hello hello' * 10**6)
+        f.write(b'hello hello' * 10 ** 6)
 
-    assert s3.cat('mybucket/myfile') == b'hello hello' * 10**6
+    assert s3.cat('mybucket/myfile') == b'hello hello' * 10 ** 6
 
 
 def test_write_fails(s3):
     with pytest.raises(ValueError):
-        s3.touch(test_bucket_name+'/temp')
-        s3.open(test_bucket_name+'/temp', 'rb').write(b'hello')
+        s3.touch(test_bucket_name + '/temp')
+        s3.open(test_bucket_name + '/temp', 'rb').write(b'hello')
     with pytest.raises(ValueError):
-        s3.open(test_bucket_name+'/temp', 'wb', block_size=10)
-    f = s3.open(test_bucket_name+'/temp', 'wb')
+        s3.open(test_bucket_name + '/temp', 'wb', block_size=10)
+    f = s3.open(test_bucket_name + '/temp', 'wb')
     f.close()
     with pytest.raises(ValueError):
         f.write(b'hello')
@@ -672,22 +760,22 @@ def test_write_fails(s3):
 
 
 def test_write_blocks(s3):
-    with s3.open(test_bucket_name+'/temp', 'wb') as f:
-        f.write(b'a' * 2*2**20)
-        assert f.buffer.tell() == 2*2**20
-        assert not(f.parts)
+    with s3.open(test_bucket_name + '/temp', 'wb') as f:
+        f.write(b'a' * 2 * 2 ** 20)
+        assert f.buffer.tell() == 2 * 2 ** 20
+        assert not (f.parts)
         f.flush()
-        assert f.buffer.tell() == 2*2**20
-        assert not(f.parts)
-        f.write(b'a' * 2*2**20)
-        f.write(b'a' * 2*2**20)
+        assert f.buffer.tell() == 2 * 2 ** 20
+        assert not (f.parts)
+        f.write(b'a' * 2 * 2 ** 20)
+        f.write(b'a' * 2 * 2 ** 20)
         assert f.mpu
         assert f.parts
-    assert s3.info(test_bucket_name+'/temp')['Size'] == 6*2**20
-    with s3.open(test_bucket_name+'/temp', 'wb', block_size=10*2**20) as f:
-        f.write(b'a' * 15*2**20)
+    assert s3.info(test_bucket_name + '/temp')['Size'] == 6 * 2 ** 20
+    with s3.open(test_bucket_name + '/temp', 'wb', block_size=10 * 2 ** 20) as f:
+        f.write(b'a' * 15 * 2 ** 20)
         assert f.buffer.tell() == 0
-    assert s3.info(test_bucket_name+'/temp')['Size'] == 15*2**20
+    assert s3.info(test_bucket_name + '/temp')['Size'] == 15 * 2 ** 20
 
 
 def test_readline(s3):
@@ -749,7 +837,7 @@ def test_readline_empty(s3):
 
 
 def test_readline_blocksize(s3):
-    data = b'ab\n' + b'a' * (10 * 2**20) + b'\nab'
+    data = b'ab\n' + b'a' * (10 * 2 ** 20) + b'\nab'
     with s3.open(a, 'wb') as f:
         f.write(data)
     with s3.open(a, 'rb') as f:
@@ -758,7 +846,7 @@ def test_readline_blocksize(s3):
         assert result == expected
 
         result = f.readline()
-        expected = b'a' * (10 * 2**20) + b'\n'
+        expected = b'a' * (10 * 2 ** 20) + b'\n'
         assert result == expected
 
         result = f.readline()
@@ -822,44 +910,44 @@ def test_writable(s3):
 
 def test_merge(s3):
     with s3.open(a, 'wb') as f:
-        f.write(b'a' * 10*2**20)
+        f.write(b'a' * 10 * 2 ** 20)
 
     with s3.open(b, 'wb') as f:
-        f.write(b'a' * 10*2**20)
-    s3.merge(test_bucket_name+'/joined', [a, b])
-    assert s3.info(test_bucket_name+'/joined')['Size'] == 2*10*2**20
+        f.write(b'a' * 10 * 2 ** 20)
+    s3.merge(test_bucket_name + '/joined', [a, b])
+    assert s3.info(test_bucket_name + '/joined')['Size'] == 2 * 10 * 2 ** 20
 
 
 def test_append(s3):
     data = text_files['nested/file1']
-    with s3.open(test_bucket_name+'/nested/file1', 'ab') as f:
-        assert f.tell() == len(data) # append, no write, small file
-    assert s3.cat(test_bucket_name+'/nested/file1') == data
-    with s3.open(test_bucket_name+'/nested/file1', 'ab') as f:
+    with s3.open(test_bucket_name + '/nested/file1', 'ab') as f:
+        assert f.tell() == len(data)  # append, no write, small file
+    assert s3.cat(test_bucket_name + '/nested/file1') == data
+    with s3.open(test_bucket_name + '/nested/file1', 'ab') as f:
         f.write(b'extra')  # append, write, small file
-    assert s3.cat(test_bucket_name+'/nested/file1') == data+b'extra'
+    assert s3.cat(test_bucket_name + '/nested/file1') == data + b'extra'
 
     with s3.open(a, 'wb') as f:
-        f.write(b'a' * 10*2**20)
+        f.write(b'a' * 10 * 2 ** 20)
     with s3.open(a, 'ab') as f:
         pass  # append, no write, big file
-    assert s3.cat(a) == b'a' * 10*2**20
+    assert s3.cat(a) == b'a' * 10 * 2 ** 20
 
     with s3.open(a, 'ab') as f:
         assert f.parts
-        assert f.tell() == 10*2**20
+        assert f.tell() == 10 * 2 ** 20
         f.write(b'extra')  # append, small write, big file
-    assert s3.cat(a) == b'a' * 10*2**20 + b'extra'
+    assert s3.cat(a) == b'a' * 10 * 2 ** 20 + b'extra'
 
     with s3.open(a, 'ab') as f:
-        assert f.tell() == 10*2**20 + 5
-        f.write(b'b' * 10*2**20)  # append, big write, big file
-        assert f.tell() == 20*2**20 + 5
-    assert s3.cat(a) == b'a' * 10*2**20 + b'extra' + b'b' *10*2**20
+        assert f.tell() == 10 * 2 ** 20 + 5
+        f.write(b'b' * 10 * 2 ** 20)  # append, big write, big file
+        assert f.tell() == 20 * 2 ** 20 + 5
+    assert s3.cat(a) == b'a' * 10 * 2 ** 20 + b'extra' + b'b' * 10 * 2 ** 20
 
 
 def test_bigger_than_block_read(s3):
-    with s3.open(test_bucket_name+'/2014-01-01.csv', 'rb', block_size=3) as f:
+    with s3.open(test_bucket_name + '/2014-01-01.csv', 'rb', block_size=3) as f:
         out = []
         while True:
             data = f.read(20)
@@ -886,7 +974,7 @@ def test_array(s3):
 
 
 def _get_s3_id(s3):
-        return id(s3.connect())
+    return id(s3.connect())
 
 
 def test_no_connection_sharing_among_processes(s3):
@@ -904,9 +992,9 @@ def test_public_file(s3):
         other_bucket_name = 's3fs_private_test'
 
         s3.touch(test_bucket_name)
-        s3.touch(test_bucket_name+'/afile')
+        s3.touch(test_bucket_name + '/afile')
         s3.touch(other_bucket_name, acl='public-read')
-        s3.touch(other_bucket_name+'/afile', acl='public-read')
+        s3.touch(other_bucket_name + '/afile', acl='public-read')
 
         s = S3FileSystem(anon=True)
         with pytest.raises((IOError, OSError)):
@@ -920,9 +1008,9 @@ def test_public_file(s3):
         assert s.ls(test_bucket_name, refresh=True)
 
         # public file in private bucket
-        with s3.open(other_bucket_name+'/see_me', 'wb', acl='public-read') as f:
+        with s3.open(other_bucket_name + '/see_me', 'wb', acl='public-read') as f:
             f.write(b'hello')
-        assert s.cat(other_bucket_name+'/see_me') == b'hello'
+        assert s.cat(other_bucket_name + '/see_me') == b'hello'
     finally:
         s3.rm(test_bucket_name, recursive=True)
         s3.rm(other_bucket_name, recursive=True)
@@ -956,7 +1044,7 @@ def test_multipart_upload_blocksize(s3):
 def test_not_fill_cache(s3):
     fname = list(files)[0]
     data = files[fname]
-    with s3.open(test_bucket_name+'/'+fname, 'rb', block_size=5,
+    with s3.open(test_bucket_name + '/' + fname, 'rb', block_size=5,
                  fill_cache=False) as f:
         f.read(6)
         assert len(f.cache) == 11
@@ -1084,12 +1172,12 @@ def test_readinto(s3):
     with s3.open('bucket/file.txt', 'wb') as fd:
         fd.write(b'Hello, World!')
 
-    contents = bytearray()
+    contents = bytearray(15)
 
     with s3.open('bucket/file.txt', 'rb') as fd:
         assert fd.readinto(contents) == 13
 
-    assert contents == b'Hello, World!'
+    assert contents.startswith(b'Hello, World!')
 
 
 def test_change_defaults_only_subsequent(monkeypatch):
@@ -1117,7 +1205,6 @@ def test_change_defaults_only_subsequent(monkeypatch):
     assert fs_default.default_block_size == 5 * (1024 ** 2)
 
 
-
 def test_passed_in_session_set_correctly(s3):
     session = boto3.session.Session()
     s3 = S3FileSystem(session=session)
@@ -1147,6 +1234,3 @@ def test_pickle_with_passed_in_session(s3):
     s3 = S3FileSystem(session=session)
     with pytest.raises(NotImplementedError):
         s3.__getstate__()
-
-
-
